from __future__ import (
    annotations,
)

from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    TYPE_CHECKING,
    AsyncIterator,
    Awaitable,
    Optional,
)
from uuid import (
    UUID,
)

from minos.common import (
    MinosSetup,
)

from ..queries import (
    _TRUE_CONDITION,
    _Condition,
    _Ordering,
)
from ..transactions import (
    TRANSACTION_CONTEXT_VAR,
    TransactionEntry,
)

if TYPE_CHECKING:
    from ..entities import (
        RootEntity,
    )


class SnapshotRepository(ABC, MinosSetup):
    """Base Snapshot class.

    The snapshot provides a direct accessor to the ``RootEntity`` instances stored as events by the event repository
    class.
    """

    async def get(self, name: str, uuid: UUID, transaction: Optional[TransactionEntry] = None, **kwargs) -> RootEntity:
        """Get a ``RootEntity`` instance from its identifier.

        :param name: Class name of the ``RootEntity``.
        :param uuid: Identifier of the ``RootEntity``.
        :param transaction: The transaction within the operation is performed. If not any value is provided, then the
            transaction is extracted from the context var. If not any transaction is being scoped then the query is
            performed to the global snapshot.
        :param kwargs: Additional named arguments.
        :return: The ``RootEntity`` instance.
        """
        if transaction is None:
            transaction = TRANSACTION_CONTEXT_VAR.get()

        await self.synchronize(**kwargs)

        return await self._get(name=name, uuid=uuid, transaction=transaction, **kwargs)

    @abstractmethod
    async def _get(self, *args, **kwargs) -> RootEntity:
        raise NotImplementedError

    def get_all(
        self,
        name: str,
        ordering: Optional[_Ordering] = None,
        limit: Optional[int] = None,
        streaming_mode: bool = False,
        transaction: Optional[TransactionEntry] = None,
        **kwargs,
    ) -> AsyncIterator[RootEntity]:
        """Get all ``RootEntity`` instances.

        :param name: Class name of the ``RootEntity``.
        :param ordering: Optional argument to return the instance with specific ordering strategy. The default behaviour
            is to retrieve them without any order pattern.
        :param limit: Optional argument to return only a subset of instances. The default behaviour is to return all the
            instances that meet the given condition.
        :param streaming_mode: If ``True`` return the values in streaming directly from the database (keep an open
            database connection), otherwise preloads the full set of values on memory and then retrieves them.
        :param transaction: The transaction within the operation is performed. If not any value is provided, then the
            transaction is extracted from the context var. If not any transaction is being scoped then the query is
            performed to the global snapshot.
        :param kwargs: Additional named arguments.
        :return: An asynchronous iterator that containing the ``RootEntity`` instances.
        """
        return self.find(
            name,
            _TRUE_CONDITION,
            ordering=ordering,
            limit=limit,
            streaming_mode=streaming_mode,
            transaction=transaction,
            **kwargs,
        )

    async def find(
        self,
        name: str,
        condition: _Condition,
        ordering: Optional[_Ordering] = None,
        limit: Optional[int] = None,
        streaming_mode: bool = False,
        transaction: Optional[TransactionEntry] = None,
        **kwargs,
    ) -> AsyncIterator[RootEntity]:
        """Find a collection of ``RootEntity`` instances based on a ``Condition``.

        :param name: Class name of the ``RootEntity``.
        :param condition: The condition that must be satisfied by the ``RootEntity`` instances.
        :param ordering: Optional argument to return the instance with specific ordering strategy. The default behaviour
            is to retrieve them without any order pattern.
        :param limit: Optional argument to return only a subset of instances. The default behaviour is to return all the
            instances that meet the given condition.
        :param streaming_mode: If ``True`` return the values in streaming directly from the database (keep an open
            database connection), otherwise preloads the full set of values on memory and then retrieves them.
        :param transaction: The transaction within the operation is performed. If not any value is provided, then the
            transaction is extracted from the context var. If not any transaction is being scoped then the query is
            performed to the global snapshot.
        :param kwargs: Additional named arguments.
        :return: An asynchronous iterator that containing the ``RootEntity`` instances.
        """
        if transaction is None:
            transaction = TRANSACTION_CONTEXT_VAR.get()

        await self.synchronize(**kwargs)

        iterable = self._find(
            name=name,
            condition=condition,
            ordering=ordering,
            limit=limit,
            streaming_mode=streaming_mode,
            transaction=transaction,
            **kwargs,
        )

        async for instance in iterable:
            yield instance

    @abstractmethod
    def _find(self, *args, **kwargs) -> AsyncIterator[RootEntity]:
        raise NotImplementedError

    def synchronize(self, **kwargs) -> Awaitable[None]:
        """Synchronize the snapshot to the latest available version.

        :param kwargs: Additional named arguments.
        :return: This method does not return anything.
        """
        return self._synchronize(**kwargs)

    @abstractmethod
    async def _synchronize(self, **kwargs) -> None:
        raise NotImplementedError



from __future__ import (
    annotations,
)

from typing import (
    Generic,
    TypeVar,
    get_args,
)

from dependency_injector.wiring import (
    Provide,
    inject,
)

from minos.common import (
    MinosConfig,
    MinosSetup,
    NotProvidedException,
)

from .entities import (
    RootEntity,
)
from .events import (
    EventRepository,
)
from .snapshots import (
    SnapshotRepository,
)
from .transactions import (
    TransactionRepository,
)

RT = TypeVar("RT", bound=RootEntity)


class Aggregate(Generic[RT], MinosSetup):
    """Base Service class"""

    transaction_repository: TransactionRepository
    event_repository: EventRepository
    snapshot_repository: SnapshotRepository

    def __init__(
        self,
        transaction_repository: TransactionRepository,
        event_repository: EventRepository,
        snapshot_repository: SnapshotRepository,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._check_root()

        self.transaction_repository = transaction_repository
        self.event_repository = event_repository
        self.snapshot_repository = snapshot_repository

    @classmethod
    def _from_config(cls, config: MinosConfig, **kwargs) -> Aggregate:
        kwargs["transaction_repository"] = cls._get_transaction_repository(**kwargs)
        kwargs["event_repository"] = cls._get_event_repository(**kwargs)
        kwargs["snapshot_repository"] = cls._get_snapshot_repository(**kwargs)
        return cls(**kwargs)

    # noinspection PyUnusedLocal
    @staticmethod
    @inject
    def _get_transaction_repository(
        transaction_repository: TransactionRepository = Provide["transaction_repository"], **kwargs
    ) -> TransactionRepository:
        if transaction_repository is None or isinstance(transaction_repository, Provide):
            raise NotProvidedException(f"A {TransactionRepository!r} object must be provided.")
        return transaction_repository

    # noinspection PyUnusedLocal
    @staticmethod
    @inject
    def _get_event_repository(
        event_repository: EventRepository = Provide["event_repository"], **kwargs
    ) -> EventRepository:
        if event_repository is None or isinstance(event_repository, Provide):
            raise NotProvidedException(f"A {EventRepository!r} object must be provided.")
        return event_repository

    # noinspection PyUnusedLocal
    @staticmethod
    @inject
    def _get_snapshot_repository(
        snapshot_repository: SnapshotRepository = Provide["snapshot_repository"], **kwargs
    ) -> SnapshotRepository:
        if snapshot_repository is None or isinstance(snapshot_repository, Provide):
            raise NotProvidedException(f"A {SnapshotRepository!r} object must be provided.")
        return snapshot_repository

    def _check_root(self):
        self.root  # If root is not valid it will raise an exception.

    @property
    def root(self) -> type[RootEntity]:
        """Get the root entity of the aggregate.

        :return: A ``RootEntity`` type.
        """
        # noinspection PyUnresolvedReferences
        bases = self.__orig_bases__
        root = get_args(bases[0])[0]
        if not isinstance(root, type) or not issubclass(root, RootEntity):
            raise TypeError(f"{type(self)!r} must contain a {RootEntity!r} as generic value.")
        return root



from __future__ import (
    annotations,
)

from typing import (
    Any,
    Optional,
)
from uuid import (
    UUID,
    uuid4,
)

from psycopg2.extras import (
    Json,
)
from psycopg2.sql import (
    SQL,
    Composable,
    Identifier,
    Literal,
    Placeholder,
)

from minos.common import (
    NULL_UUID,
    AvroDataEncoder,
)

from ...contextvars import (
    IS_REPOSITORY_SERIALIZATION_CONTEXT_VAR,
)
from ...queries import (
    _FALSE_CONDITION,
    _AndCondition,
    _ComposedCondition,
    _Condition,
    _EqualCondition,
    _FalseCondition,
    _GreaterCondition,
    _GreaterEqualCondition,
    _InCondition,
    _LikeCondition,
    _LowerCondition,
    _LowerEqualCondition,
    _NotCondition,
    _NotEqualCondition,
    _OrCondition,
    _Ordering,
    _SimpleCondition,
    _TrueCondition,
)


class PostgreSqlSnapshotQueryBuilder:
    """PostgreSQL Snapshot Query Builder class.

    This class build postgres-compatible database queries over fields based on a condition, ordering, etc.
    """

    def __init__(
        self,
        name: str,
        condition: _Condition,
        ordering: Optional[_Ordering] = None,
        limit: Optional[int] = None,
        transaction_uuids: tuple[UUID, ...] = (NULL_UUID,),
        exclude_deleted: bool = False,
    ):
        self.name = name
        self.condition = condition
        self.ordering = ordering
        self.limit = limit
        self.transaction_uuids = transaction_uuids
        self.exclude_deleted = exclude_deleted
        self._parameters = None

    def build(self) -> tuple[Composable, dict[str, Any]]:
        """Build a query.

        :return: A tuple in which the first value is the sql sentence and the second one is a dictionary containing the
            query parameters.
        """
        self._parameters = dict()

        token = IS_REPOSITORY_SERIALIZATION_CONTEXT_VAR.set(True)
        try:
            query = self._build()
        finally:
            IS_REPOSITORY_SERIALIZATION_CONTEXT_VAR.reset(token)

        parameters = self._parameters

        return query, parameters

    def _build(self) -> Composable:
        self._parameters["name"] = self.name

        query = SQL(" WHERE ").join([self._build_select_from(), self._build_condition(self.condition)])

        if self.exclude_deleted:
            query = SQL(" AND ").join([query, _EXCLUDE_DELETED_CONDITION])

        if self.ordering is not None:
            query = SQL(" ").join([query, self._build_ordering(self.ordering)])

        if self.limit is not None:
            query = SQL(" ").join([query, self._build_limit(self.limit)])

        return query

    def _build_select_from(self) -> Composable:
        from_query_parts = list()
        for index, transaction_uuid in enumerate(self.transaction_uuids, start=1):
            name = f"transaction_uuid_{index}"
            self._parameters[name] = transaction_uuid

            from_query_parts.append(
                _SELECT_TRANSACTION_CHUNK.format(index=Literal(index), transaction_uuid=Placeholder(name))
            )

        from_query = SQL(" UNION ALL ").join(from_query_parts)

        query = _SELECT_ENTRIES_QUERY.format(from_parts=from_query)
        return query

    def _build_condition(self, condition: _Condition) -> Composable:
        if isinstance(condition, _NotCondition):
            return self._build_condition_not(condition)
        if isinstance(condition, _ComposedCondition):
            return self._build_condition_composed(condition)
        if isinstance(condition, _TrueCondition):
            return SQL("TRUE")
        if isinstance(condition, _FalseCondition):
            return SQL("FALSE")
        if isinstance(condition, _LikeCondition):
            return self._build_condition_like(condition)
        if isinstance(condition, _SimpleCondition):
            return self._build_condition_simple(condition)

        raise ValueError(f"Given condition is not supported. Obtained: {condition}")

    def _build_condition_not(self, condition: _NotCondition) -> Composable:
        return SQL("(NOT {})").format(self._build_condition(condition.inner))

    def _build_condition_composed(self, condition: _ComposedCondition) -> Composable:
        # noinspection PyTypeChecker
        operator = _COMPOSED_MAPPER[type(condition)]
        parts = (self._build_condition(c) for c in condition)
        return SQL("({composed})").format(composed=operator.join(parts))

    def _build_condition_simple(self, condition: _SimpleCondition) -> Composable:
        field = condition.field
        # noinspection PyTypeChecker
        operator = _SIMPLE_MAPPER[type(condition)]

        parameter = AvroDataEncoder(condition.parameter).build()
        if isinstance(parameter, list):
            if not len(parameter):
                return self._build_condition(_FALSE_CONDITION)
            parameter = tuple(parameter)

        if field in _FIXED_FIELDS_MAPPER:
            name = self.generate_random_str()
            self._parameters[name] = parameter

            field = _FIXED_FIELDS_MAPPER[field]
            name = Placeholder(name)
            return SQL("({field} {operator} {name})").format(field=field, operator=operator, name=name)
        else:
            name = self.generate_random_str()
            self._parameters[name] = Json(parameter)

            field = Literal("{{{}}}".format(field.replace(".", ",")))
            name = Placeholder(name)
            return SQL("(data#>{field} {operator} {name}::jsonb)").format(field=field, operator=operator, name=name)

    def _build_condition_like(self, condition: _SimpleCondition) -> Composable:
        field = condition.field

        parameter = AvroDataEncoder(condition.parameter).build()

        if field in _FIXED_FIELDS_MAPPER:
            name = self.generate_random_str()
            self._parameters[name] = parameter

            field = _FIXED_FIELDS_MAPPER[field]
            name = Placeholder(name)
            return SQL("({field}::text LIKE {name})").format(field=field, name=name)
        else:
            name = self.generate_random_str()
            self._parameters[name] = parameter

            field = Literal("{{{}}}".format(field.replace(".", ",")))
            name = Placeholder(name)
            return SQL("(data#>>{field} LIKE {name})").format(field=field, name=name)

    @staticmethod
    def _build_ordering(ordering: _Ordering) -> Composable:
        field = ordering.by
        direction = _ORDERING_MAPPER[ordering.reverse]

        if field in _FIXED_FIELDS_MAPPER:
            field = Identifier(field)
            order_by = SQL("ORDER BY {field} {direction}").format(field=field, direction=direction)
        else:
            field = Literal("{{{}}}".format(field.replace(".", ",")))
            order_by = SQL("ORDER BY data#>{field} {direction}").format(field=field, direction=direction)

        return order_by

    @staticmethod
    def _build_limit(value: int) -> Composable:
        limit = SQL("LIMIT {limit}").format(limit=Literal(value))
        return limit

    @staticmethod
    def generate_random_str() -> str:
        """Generate a random string

        :return: A random string value.
        """
        return str(uuid4())


_COMPOSED_MAPPER = {_AndCondition: SQL(" AND "), _OrCondition: SQL(" OR ")}

_SIMPLE_MAPPER = {
    _LowerCondition: SQL("<"),
    _LowerEqualCondition: SQL("<="),
    _GreaterCondition: SQL(">"),
    _GreaterEqualCondition: SQL(">="),
    _EqualCondition: SQL("="),
    _NotEqualCondition: SQL("<>"),
    _InCondition: SQL("IN"),
}

_FIXED_FIELDS_MAPPER = {
    "uuid": Identifier("uuid"),
    "version": Identifier("version"),
    "created_at": Identifier("created_at"),
    "updated_at": Identifier("updated_at"),
}

_ORDERING_MAPPER = {
    True: SQL("DESC"),
    False: SQL("ASC"),
}

_SELECT_ENTRIES_QUERY = SQL(
    "SELECT "
    "   t2.uuid, "
    "   t2.name, "
    "   t2.version, "
    "   t2.schema, "
    "   t2.data, "
    "   t2.created_at, "
    "   t2.updated_at, "
    "   t2.transaction_uuid "
    "FROM ("
    "   SELECT DISTINCT ON (uuid) t1.* "
    "   FROM ( {from_parts} ) AS t1 "
    "   ORDER BY uuid, transaction_index DESC "
    ") AS t2"
)

_SELECT_TRANSACTION_CHUNK = SQL(
    "SELECT {index} AS transaction_index, * "
    "FROM snapshot "
    "WHERE name = %(name)s AND transaction_uuid = {transaction_uuid} "
)

_EXCLUDE_DELETED_CONDITION = SQL("(data IS NOT NULL)")



from .contextvars import (
    TRANSACTION_CONTEXT_VAR,
)
from .entries import (
    TransactionEntry,
    TransactionStatus,
)
from .repositories import (
    InMemoryTransactionRepository,
    PostgreSqlTransactionRepository,
    TransactionRepository,
)
from .services import (
    TransactionService,
)



from __future__ import (
    annotations,
)

from datetime import (
    datetime,
)
from typing import (
    AsyncIterator,
    Optional,
)
from uuid import (
    UUID,
)

from minos.common import (
    MinosConfig,
    PostgreSqlMinosDatabase,
)

from ...exceptions import (
    TransactionRepositoryConflictException,
)
from ..entries import (
    TransactionEntry,
)
from .abc import (
    TransactionRepository,
)


class PostgreSqlTransactionRepository(PostgreSqlMinosDatabase, TransactionRepository):
    """PostgreSql Transaction Repository class."""

    @classmethod
    def _from_config(cls, *args, config: MinosConfig, **kwargs) -> Optional[PostgreSqlTransactionRepository]:
        return cls(*args, **config.repository._asdict(), **kwargs)

    async def _setup(self):
        await self.submit_query('CREATE EXTENSION IF NOT EXISTS "uuid-ossp";', lock="uuid-ossp")

        await self.submit_query(_CREATE_TRANSACTION_STATUS_ENUM_QUERY, lock=hash("aggregate_transaction_enum"))
        await self.submit_query(_CREATE_TRANSACTION_TABLE_QUERY, lock=hash("aggregate_transaction"))

    async def _submit(self, transaction: TransactionEntry) -> TransactionEntry:
        params = {
            "uuid": transaction.uuid,
            "destination_uuid": transaction.destination_uuid,
            "status": transaction.status,
            "event_offset": transaction.event_offset,
        }
        try:
            updated_at = await self.submit_query_and_fetchone(
                _INSERT_TRANSACTIONS_VALUES_QUERY, params, lock=transaction.uuid.int & (1 << 32) - 1
            )
        except StopAsyncIteration:
            raise TransactionRepositoryConflictException(
                f"{transaction!r} status is invalid respect to the previous one."
            )
        transaction.updated_at = updated_at
        return transaction

    async def _select(self, **kwargs) -> AsyncIterator[TransactionEntry]:
        query = self._build_select_query(**kwargs)
        async for row in self.submit_query_and_iter(query, kwargs, **kwargs):
            yield TransactionEntry(*row, transaction_repository=self)

    # noinspection PyUnusedLocal
    @staticmethod
    def _build_select_query(
        uuid: Optional[UUID] = None,
        uuid_ne: Optional[UUID] = None,
        uuid_in: Optional[tuple[UUID]] = None,
        destination_uuid: Optional[UUID] = None,
        status: Optional[str] = None,
        status_in: Optional[tuple[str]] = None,
        event_offset: Optional[int] = None,
        event_offset_lt: Optional[int] = None,
        event_offset_gt: Optional[int] = None,
        event_offset_le: Optional[int] = None,
        event_offset_ge: Optional[int] = None,
        updated_at: Optional[datetime] = None,
        updated_at_lt: Optional[datetime] = None,
        updated_at_gt: Optional[datetime] = None,
        updated_at_le: Optional[datetime] = None,
        updated_at_ge: Optional[datetime] = None,
        **kwargs,
    ) -> str:
        conditions = list()

        if uuid is not None:
            conditions.append("uuid = %(uuid)s")
        if uuid_ne is not None:
            conditions.append("uuid <> %(uuid_ne)s")
        if uuid_in is not None:
            conditions.append("uuid IN %(uuid_in)s")
        if destination_uuid is not None:
            conditions.append("destination_uuid = %(destination_uuid)s")
        if status is not None:
            conditions.append("status = %(status)s")
        if status_in is not None:
            conditions.append("status IN %(status_in)s")
        if event_offset is not None:
            conditions.append("event_offset = %(event_offset)s")
        if event_offset_lt is not None:
            conditions.append("event_offset < %(event_offset_lt)s")
        if event_offset_gt is not None:
            conditions.append("event_offset > %(event_offset_gt)s")
        if event_offset_le is not None:
            conditions.append("event_offset <= %(event_offset_le)s")
        if event_offset_ge is not None:
            conditions.append("event_offset >= %(event_offset_ge)s")
        if updated_at is not None:
            conditions.append("updated_at = %(updated_at)s")
        if updated_at_lt is not None:
            conditions.append("updated_at < %(updated_at_lt)s")
        if updated_at_gt is not None:
            conditions.append("updated_at > %(updated_at_gt)s")
        if updated_at_le is not None:
            conditions.append("updated_at <= %(updated_at_le)s")
        if updated_at_ge is not None:
            conditions.append("updated_at >= %(updated_at_ge)s")

        if not conditions:
            return f"{_SELECT_ALL_TRANSACTIONS_QUERY} ORDER BY event_offset;"

        return f"{_SELECT_ALL_TRANSACTIONS_QUERY} WHERE {' AND '.join(conditions)} ORDER BY event_offset;"


_CREATE_TRANSACTION_STATUS_ENUM_QUERY = """
DO
$$
    BEGIN
        IF NOT EXISTS(SELECT *
                      FROM pg_type typ
                               INNER JOIN pg_namespace nsp
                                          ON nsp.oid = typ.typnamespace
                      WHERE nsp.nspname = current_schema()
                        AND typ.typname = 'transaction_status') THEN
            CREATE TYPE transaction_status AS ENUM (
                'pending', 'reserving', 'reserved', 'committing', 'committed', 'rejected'
            );
        END IF;
    END;
$$
LANGUAGE plpgsql;
""".strip()

_CREATE_TRANSACTION_TABLE_QUERY = """
CREATE TABLE IF NOT EXISTS aggregate_transaction (
    uuid UUID PRIMARY KEY,
    destination_uuid UUID NOT NULL,
    status TRANSACTION_STATUS NOT NULL,
    event_offset INTEGER,
    updated_at TIMESTAMPTZ NOT NULL DEFAULT NOW()
);
""".strip()

_INSERT_TRANSACTIONS_VALUES_QUERY = """
INSERT INTO aggregate_transaction (uuid, destination_uuid, status, event_offset)
VALUES (%(uuid)s, %(destination_uuid)s, %(status)s, %(event_offset)s)
ON CONFLICT (uuid)
DO
   UPDATE SET status = %(status)s, event_offset = %(event_offset)s, updated_at = NOW()
WHERE (aggregate_transaction.destination_uuid = %(destination_uuid)s)
  AND (NOT (aggregate_transaction.status = 'pending' AND %(status)s NOT IN ('pending', 'reserving', 'rejected')))
  AND (NOT (aggregate_transaction.status = 'reserving' AND %(status)s NOT IN ('reserved', 'rejected')))
  AND (NOT (aggregate_transaction.status = 'reserved' AND %(status)s NOT IN ('committing', 'rejected')))
  AND (NOT (aggregate_transaction.status = 'committing' AND %(status)s NOT IN ('committed')))
  AND (NOT (aggregate_transaction.status = 'committed'))
  AND (NOT (aggregate_transaction.status = 'rejected'))
RETURNING updated_at;
""".strip()

_SELECT_ALL_TRANSACTIONS_QUERY = """
SELECT uuid, status, event_offset, destination_uuid, updated_at
FROM aggregate_transaction
""".strip()



from __future__ import (
    annotations,
)

from contextvars import (
    ContextVar,
)
from typing import (
    TYPE_CHECKING,
    Final,
    Optional,
)

if TYPE_CHECKING:
    from .entries import (
        TransactionEntry,
    )

TRANSACTION_CONTEXT_VAR: Final[ContextVar[Optional[TransactionEntry]]] = ContextVar("transaction", default=None)



from asyncio import (
    gather,
)
from collections.abc import (
    Iterable,
)
from itertools import (
    chain,
)
from typing import (
    Any,
)
from uuid import (
    UUID,
)

from dependency_injector.wiring import (
    Provide,
    inject,
)

from minos.common import (
    Model,
)
from minos.networks import (
    BrokerClient,
    BrokerClientPool,
    BrokerMessageV1,
    BrokerMessageV1Payload,
)

from .extractors import (
    RefExtractor,
)
from .injectors import (
    RefInjector,
)


class RefResolver:
    """Ref Resolver class."""

    # noinspection PyUnusedLocal
    @inject
    def __init__(self, broker_pool: BrokerClientPool = Provide["broker_pool"], **kwargs):
        self.broker_pool = broker_pool

    # noinspection PyUnusedLocal
    async def resolve(self, data: Any, **kwargs) -> Any:
        """Resolve Ref instances.

        :param data: The data to be resolved.
        :param kwargs: Additional named arguments.
        :return: The data instance with model references already resolved.
        """
        missing = RefExtractor(data).build()

        if not len(missing):
            return data

        recovered = await self._query(missing)

        return RefInjector(data, recovered).build()

    async def _query(self, references: dict[str, set[UUID]]) -> dict[UUID, Model]:
        async with self.broker_pool.acquire() as broker:
            futures = (
                broker.send(
                    BrokerMessageV1(f"_Get{simplified_name}Snapshots", BrokerMessageV1Payload({"uuids": uuids}))
                )
                for simplified_name, uuids in references.items()
            )
            await gather(*futures)

            return {model.uuid: model for model in await self._get_response(broker, len(references))}

    @staticmethod
    async def _get_response(broker: BrokerClient, count: int, **kwargs) -> Iterable[Model]:
        messages = [message async for message in broker.receive_many(count, **kwargs)]
        return chain(*(message.content for message in messages))



from abc import (
    ABC,
    abstractmethod,
)
from datetime import (
    datetime,
)
from typing import (
    AsyncIterator,
    Optional,
)
from uuid import (
    UUID,
)

from dependency_injector.wiring import (
    Provide,
    inject,
)

from minos.common import (
    Lock,
    MinosPool,
    MinosSetup,
    NotProvidedException,
)

from ...exceptions import (
    TransactionNotFoundException,
)
from ..entries import (
    TransactionEntry,
    TransactionStatus,
)


class TransactionRepository(ABC, MinosSetup):
    """Transaction Repository base class."""

    @inject
    def __init__(self, lock_pool: MinosPool[Lock] = Provide["lock_pool"], *args, **kwargs):
        super().__init__(*args, **kwargs)

        if lock_pool is None or isinstance(lock_pool, Provide):
            raise NotProvidedException("A lock pool instance is required.")

        self._lock_pool = lock_pool

    async def submit(self, transaction: TransactionEntry) -> TransactionEntry:
        """Submit a new or updated transaction to store it on the repository.

        :param transaction: The transaction to be stored.
        :return: This method does not return anything.
        """
        return await self._submit(transaction)

    @abstractmethod
    async def _submit(self, transaction: TransactionEntry) -> TransactionEntry:
        raise NotImplementedError

    # noinspection PyUnusedLocal
    async def get(self, uuid: UUID, **kwargs) -> TransactionEntry:
        """Get a ``TransactionEntry`` from its identifier.

        :param uuid: Identifier of the ``RootEntity``.
        :param kwargs: Additional named arguments.
        :return: The ``TransactionEntry`` instance.
        """
        try:
            return await self.select(uuid=uuid).__anext__()
        except StopAsyncIteration:
            raise TransactionNotFoundException(f"Transaction identified by {uuid!r} does not exist.")

    async def select(
        self,
        uuid: Optional[UUID] = None,
        uuid_ne: Optional[UUID] = None,
        uuid_in: Optional[tuple[UUID, ...]] = None,
        destination_uuid: Optional[UUID] = None,
        status: Optional[TransactionStatus] = None,
        status_in: Optional[tuple[str, ...]] = None,
        event_offset: Optional[int] = None,
        event_offset_lt: Optional[int] = None,
        event_offset_gt: Optional[int] = None,
        event_offset_le: Optional[int] = None,
        event_offset_ge: Optional[int] = None,
        updated_at: Optional[datetime] = None,
        updated_at_lt: Optional[datetime] = None,
        updated_at_gt: Optional[datetime] = None,
        updated_at_le: Optional[datetime] = None,
        updated_at_ge: Optional[datetime] = None,
        **kwargs,
    ) -> AsyncIterator[TransactionEntry]:
        """Get a transaction from the repository.

        :param uuid: Transaction identifier equal to the given value.
        :param uuid_ne: Transaction identifier not equal to the given value
        :param uuid_in: Transaction identifier within the given values.
        :param destination_uuid: Destination Transaction identifier equal to the given value.
        :param status: Transaction status equal to the given value.
        :param status_in: Transaction status within the given values
        :param event_offset: Event offset equal to the given value.
        :param event_offset_lt: Event Offset lower than the given value
        :param event_offset_gt: Event Offset greater than the given value
        :param event_offset_le: Event Offset lower or equal to the given value
        :param event_offset_ge: Event Offset greater or equal to the given value
        :param updated_at: Updated at equal to the given value.
        :param updated_at_lt: Updated at lower than the given value.
        :param updated_at_gt: Updated at greater than the given value.
        :param updated_at_le: Updated at lower or equal to the given value.
        :param updated_at_ge: Updated at greater or equal to the given value.
        :param kwargs: Additional named arguments.
        :return: An asynchronous iterator.
        """
        generator = self._select(
            uuid=uuid,
            uuid_ne=uuid_ne,
            uuid_in=uuid_in,
            destination_uuid=destination_uuid,
            status=status,
            status_in=status_in,
            event_offset=event_offset,
            event_offset_lt=event_offset_lt,
            event_offset_gt=event_offset_gt,
            event_offset_le=event_offset_le,
            event_offset_ge=event_offset_ge,
            updated_at=updated_at,
            updated_at_lt=updated_at_lt,
            updated_at_gt=updated_at_gt,
            updated_at_le=updated_at_le,
            updated_at_ge=updated_at_ge,
            **kwargs,
        )
        # noinspection PyTypeChecker
        async for entry in generator:
            yield entry

    @abstractmethod
    async def _select(self, **kwargs) -> AsyncIterator[TransactionEntry]:
        raise NotImplementedError

    def write_lock(self) -> Lock:
        """Get write lock.

        :return: An asynchronous context manager.
        """
        return self._lock_pool.acquire("aggregate_transaction_write_lock")



from .entries import (
    EventEntry,
)
from .fields import (
    FieldDiff,
    FieldDiffContainer,
    IncrementalFieldDiff,
)
from .models import (
    Event,
)
from .repositories import (
    EventRepository,
    InMemoryEventRepository,
    PostgreSqlEventRepository,
)



from __future__ import (
    annotations,
)

from typing import (
    AsyncIterator,
    Optional,
)
from uuid import (
    UUID,
)

from psycopg2 import (
    IntegrityError,
)
from psycopg2.sql import (
    SQL,
    Composable,
    Literal,
    Placeholder,
)

from minos.common import (
    NULL_UUID,
    MinosConfig,
    PostgreSqlMinosDatabase,
)

from ...exceptions import (
    EventRepositoryConflictException,
)
from ..entries import (
    EventEntry,
)
from .abc import (
    EventRepository,
)


class PostgreSqlEventRepository(PostgreSqlMinosDatabase, EventRepository):
    """PostgreSQL-based implementation of the event repository class in ``Minos``."""

    @classmethod
    def _from_config(cls, *args, config: MinosConfig, **kwargs) -> Optional[EventRepository]:
        return cls(*args, **config.repository._asdict(), **kwargs)

    async def _setup(self):
        """Setup miscellaneous repository thing.

        In the PostgreSQL case, configures the needed table to be used to store the data.

        :return: This method does not return anything.
        """
        await self.submit_query('CREATE EXTENSION IF NOT EXISTS "uuid-ossp";', lock="uuid-ossp")

        await self.submit_query(_CREATE_ACTION_ENUM_QUERY, lock="aggregate_event")
        await self.submit_query(_CREATE_TABLE_QUERY, lock="aggregate_event")

    async def _submit(self, entry: EventEntry, **kwargs) -> EventEntry:
        lock = None
        if entry.uuid != NULL_UUID:
            lock = entry.uuid.int & (1 << 32) - 1

        query, params = await self._build_query(entry)

        try:
            response = await self.submit_query_and_fetchone(query, params, lock=lock)
        except IntegrityError:
            raise EventRepositoryConflictException(
                f"{entry!r} could not be submitted due to a key (uuid, version, transaction) collision",
                await self.offset,
            )

        entry.id, entry.uuid, entry.version, entry.created_at = response
        return entry

    async def _build_query(self, entry: EventEntry) -> tuple[Composable, dict[str, UUID]]:
        if entry.transaction_uuid != NULL_UUID:
            transaction = await self._transaction_repository.get(uuid=entry.transaction_uuid)
            transaction_uuids = await transaction.uuids
        else:
            transaction_uuids = (NULL_UUID,)

        from_query_parts = list()
        parameters = dict()
        for index, transaction_uuid in enumerate(transaction_uuids, start=1):
            name = f"transaction_uuid_{index}"
            parameters[name] = transaction_uuid

            from_query_parts.append(
                _SELECT_TRANSACTION_CHUNK.format(index=Literal(index), transaction_uuid=Placeholder(name))
            )

        from_query = SQL(" UNION ALL ").join(from_query_parts)

        query = _INSERT_VALUES_QUERY.format(from_parts=from_query)

        return query, parameters | entry.as_raw()

    async def _select(self, **kwargs) -> AsyncIterator[EventEntry]:
        query = self._build_select_query(**kwargs)
        async for row in self.submit_query_and_iter(query, kwargs, **kwargs):
            yield EventEntry(*row)

    # noinspection PyUnusedLocal
    @staticmethod
    def _build_select_query(
        uuid: Optional[UUID] = None,
        name: Optional[str] = None,
        version: Optional[int] = None,
        version_lt: Optional[int] = None,
        version_gt: Optional[int] = None,
        version_le: Optional[int] = None,
        version_ge: Optional[int] = None,
        id: Optional[int] = None,
        id_lt: Optional[int] = None,
        id_gt: Optional[int] = None,
        id_le: Optional[int] = None,
        id_ge: Optional[int] = None,
        transaction_uuid: Optional[UUID] = None,
        transaction_uuid_ne: Optional[UUID] = None,
        transaction_uuid_in: Optional[tuple[UUID, ...]] = None,
        **kwargs,
    ) -> str:
        conditions = list()

        if uuid is not None:
            conditions.append("uuid = %(uuid)s")
        if name is not None:
            conditions.append("name = %(name)s")
        if version is not None:
            conditions.append("version = %(version)s")
        if version_lt is not None:
            conditions.append("version < %(version_lt)s")
        if version_gt is not None:
            conditions.append("version > %(version_gt)s")
        if version_le is not None:
            conditions.append("version <= %(version_le)s")
        if version_ge is not None:
            conditions.append("version >= %(version_ge)s")
        if id is not None:
            conditions.append("id = %(id)s")
        if id_lt is not None:
            conditions.append("id < %(id_lt)s")
        if id_gt is not None:
            conditions.append("id > %(id_gt)s")
        if id_le is not None:
            conditions.append("id <= %(id_le)s")
        if id_ge is not None:
            conditions.append("id >= %(id_ge)s")
        if transaction_uuid is not None:
            conditions.append("transaction_uuid = %(transaction_uuid)s")
        if transaction_uuid_ne is not None:
            conditions.append("transaction_uuid <> %(transaction_uuid_ne)s")
        if transaction_uuid_in is not None:
            conditions.append("transaction_uuid IN %(transaction_uuid_in)s")

        if not conditions:
            return f"{_SELECT_ALL_ENTRIES_QUERY} ORDER BY id;"

        return f"{_SELECT_ALL_ENTRIES_QUERY} WHERE {' AND '.join(conditions)} ORDER BY id;"

    @property
    async def _offset(self) -> int:
        return (await self.submit_query_and_fetchone(_SELECT_MAX_ID_QUERY))[0] or 0


_CREATE_ACTION_ENUM_QUERY = """
DO
$$
    BEGIN
        IF NOT EXISTS(SELECT *
                      FROM pg_type typ
                               INNER JOIN pg_namespace nsp
                                          ON nsp.oid = typ.typnamespace
                      WHERE nsp.nspname = current_schema()
                        AND typ.typname = 'action_type') THEN
            CREATE TYPE action_type AS ENUM ('create', 'update', 'delete');
        END IF;
    END;
$$
LANGUAGE plpgsql;
""".strip()

_CREATE_TABLE_QUERY = """
CREATE TABLE IF NOT EXISTS aggregate_event (
    id BIGSERIAL PRIMARY KEY,
    action ACTION_TYPE NOT NULL,
    uuid UUID NOT NULL,
    name TEXT NOT NULL,
    version INT NOT NULL,
    data BYTEA NOT NULL,
    created_at TIMESTAMPTZ NOT NULL,
    transaction_uuid UUID NOT NULL DEFAULT uuid_nil(),
    UNIQUE (uuid, version, transaction_uuid)
);
""".strip()

_INSERT_VALUES_QUERY = SQL(
    """
INSERT INTO aggregate_event (id, action, uuid, name, version, data, created_at, transaction_uuid)
VALUES (
    default,
    %(action)s,
    CASE %(uuid)s WHEN uuid_nil() THEN uuid_generate_v4() ELSE %(uuid)s END,
    %(name)s,
    (
        SELECT (CASE WHEN %(version)s IS NULL THEN 1 + COALESCE(MAX(t2.version), 0) ELSE %(version)s END)
        FROM (
                 SELECT DISTINCT ON (t1.uuid) t1.version
                 FROM ( {from_parts} ) AS t1
                 ORDER BY t1.uuid, t1.transaction_index DESC
        ) AS t2
    ),
    %(data)s,
    (CASE WHEN %(created_at)s IS NULL THEN NOW() ELSE %(created_at)s END),
    %(transaction_uuid)s
)
RETURNING id, uuid, version, created_at;
    """
)

_SELECT_TRANSACTION_CHUNK = SQL(
    """
SELECT {index} AS transaction_index, uuid, MAX(version) AS version
FROM aggregate_event
WHERE uuid = %(uuid)s AND transaction_uuid = {transaction_uuid}
GROUP BY uuid
    """
)

_SELECT_ALL_ENTRIES_QUERY = """
SELECT uuid, name, version, data, id, action, created_at, transaction_uuid
FROM aggregate_event
""".strip()

_SELECT_MAX_ID_QUERY = "SELECT MAX(id) FROM aggregate_event;".strip()



from datetime import (
    datetime,
)
from typing import (
    AsyncIterator,
    Optional,
)
from uuid import (
    UUID,
)

from minos.common import (
    current_datetime,
)

from ...exceptions import (
    TransactionRepositoryConflictException,
)
from ..entries import (
    TransactionEntry,
)
from ..entries import TransactionStatus as s
from .abc import (
    TransactionRepository,
)


class InMemoryTransactionRepository(TransactionRepository):
    """In Memory Transaction Repository class."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._storage = dict()

    async def _submit(self, transaction: TransactionEntry) -> TransactionEntry:
        transaction.updated_at = current_datetime()

        if transaction.uuid in self._storage:
            status = self._storage[transaction.uuid].status
            if (
                (status == s.PENDING and transaction.status not in (s.PENDING, s.RESERVING, s.REJECTED))
                or (status == s.RESERVING and transaction.status not in (s.RESERVED, s.REJECTED))
                or (status == s.RESERVED and transaction.status not in (s.COMMITTING, s.REJECTED))
                or (status == s.COMMITTING and transaction.status not in (s.COMMITTED,))
                or (status == s.COMMITTED)
                or (status == s.REJECTED)
            ):
                raise TransactionRepositoryConflictException(
                    f"{transaction!r} status is invalid respect to the previous one."
                )

        self._storage[transaction.uuid] = TransactionEntry(
            uuid=transaction.uuid,
            destination_uuid=transaction.destination_uuid,
            status=transaction.status,
            event_offset=transaction.event_offset,
            updated_at=transaction.updated_at,
            event_repository=transaction._event_repository,
            transaction_repository=transaction._transaction_repository,
        )

        return transaction

    async def _select(
        self,
        uuid: Optional[UUID] = None,
        uuid_ne: Optional[UUID] = None,
        uuid_in: Optional[tuple[UUID, ...]] = None,
        destination_uuid: Optional[UUID] = None,
        status: Optional[s] = None,
        status_in: Optional[tuple[str, ...]] = None,
        event_offset: Optional[int] = None,
        event_offset_lt: Optional[int] = None,
        event_offset_gt: Optional[int] = None,
        event_offset_le: Optional[int] = None,
        event_offset_ge: Optional[int] = None,
        updated_at: Optional[datetime] = None,
        updated_at_lt: Optional[datetime] = None,
        updated_at_gt: Optional[datetime] = None,
        updated_at_le: Optional[datetime] = None,
        updated_at_ge: Optional[datetime] = None,
        **kwargs,
    ) -> AsyncIterator[TransactionEntry]:

        # noinspection DuplicatedCode
        def _fn_filter(transaction: TransactionEntry) -> bool:
            if uuid is not None and uuid != transaction.uuid:
                return False
            if uuid_ne is not None and uuid_ne == transaction.uuid:
                return False
            if uuid_in is not None and transaction.uuid not in uuid_in:
                return False
            if destination_uuid is not None and destination_uuid != transaction.destination_uuid:
                return False
            if status is not None and status != transaction.status:
                return False
            if status_in is not None and transaction.status not in status_in:
                return False
            if event_offset is not None and event_offset != transaction.event_offset:
                return False
            if event_offset_lt is not None and event_offset_lt <= transaction.event_offset:
                return False
            if event_offset_gt is not None and event_offset_gt >= transaction.event_offset:
                return False
            if event_offset_le is not None and event_offset_le < transaction.event_offset:
                return False
            if event_offset_ge is not None and event_offset_ge > transaction.event_offset:
                return False
            if updated_at is not None and updated_at != transaction.updated_at:
                return False
            if updated_at_lt is not None and updated_at_lt <= transaction.updated_at:
                return False
            if updated_at_gt is not None and updated_at_gt >= transaction.updated_at:
                return False
            if updated_at_le is not None and updated_at_le < transaction.updated_at:
                return False
            if updated_at_ge is not None and updated_at_ge > transaction.updated_at:
                return False
            return True

        iterable = iter(self._storage.values())
        iterable = filter(_fn_filter, iterable)
        for item in iterable:
            yield item
